# -*- coding: utf-8 -*-
"""
ELM abstract class for API calls
"""
from abc import ABC
import os
import numpy as np
import asyncio
import aiohttp
import openai
import requests
import tiktoken
import time
import logging


logger = logging.getLogger(__name__)


class ApiBase(ABC):
    """Class to parse text from a PDF document."""

    DEFAULT_MODEL = 'gpt-3.5-turbo'
    """Default model to do pdf text cleaning."""

    EMBEDDING_MODEL = 'text-embedding-ada-002'
    """Default model to do text embeddings."""

    EMBEDDING_URL = 'https://api.openai.com/v1/embeddings'
    """OpenAI embedding API URL"""

    URL = 'https://api.openai.com/v1/chat/completions'
    """OpenAI API URL to be used with environment variable OPENAI_API_KEY. Use
    an Azure API endpoint to trigger Azure usage along with environment
    variables AZURE_OPENAI_KEY, AZURE_OPENAI_VERSION, and
    AZURE_OPENAI_ENDPOINT"""

    HEADERS = {"Content-Type": "application/json",
               "Authorization": f"Bearer {openai.api_key}",
               "api-key": f"{openai.api_key}",
               }
    """OpenAI API Headers"""

    MODEL_ROLE = "You are a research assistant that answers questions."
    """High level model role"""

    TOKENIZER_ALIASES = {'gpt-35-turbo': 'gpt-3.5-turbo',
                         'gpt-4-32k': 'gpt-4-32k-0314',
                         'llmev-gpt-4-32k': 'gpt-4-32k-0314',
                         'wetosa-gpt-4': 'gpt-4',
                         'wetosa-gpt-4-standard': 'gpt-4',
                         'wetosa-gpt-4o': 'gpt-4o',
                         }
    """Optional mappings for unusual Azure names to tiktoken/openai names."""

    TOKENIZER_PATTERNS = ('gpt-4o', 'gpt-4-32k', 'gpt-4')
    """Order-prioritized list of model sub-strings to look for in model name
    to send to tokenizer. As an alternative to alias lookup, this will use the
    tokenizer pattern if found in the model string"""

    def __init__(self, model=None):
        """
        Parameters
        ----------
        model : None | str
            Optional specification of OpenAI model to use. Default is
            cls.DEFAULT_MODEL
        """
        self.model = model or self.DEFAULT_MODEL
        self.api_queue = None
        self.messages = []
        self.clear()

        if 'openai.azure.com' in self.URL.lower():
            key = os.environ.get("AZURE_OPENAI_KEY")
            version = os.environ.get("AZURE_OPENAI_VERSION")
            endpoint = os.environ.get("AZURE_OPENAI_ENDPOINT")
            assert key is not None, "Must set AZURE_OPENAI_KEY!"
            assert version is not None, "Must set AZURE_OPENAI_VERSION!"
            assert endpoint is not None, "Must set AZURE_OPENAI_ENDPOINT!"
            self._client = openai.AzureOpenAI(api_key=key,
                                              api_version=version,
                                              azure_endpoint=endpoint)
        else:
            key = os.environ.get("OPENAI_API_KEY")
            assert key is not None, "Must set OPENAI_API_KEY!"
            self._client = openai.OpenAI(api_key=key)

    @property
    def all_messages_txt(self):
        """Get a string printout of the full conversation with the LLM

        Returns
        -------
        str
        """
        messages = [f"{msg['role'].upper()}: {msg['content']}"
                    for msg in self.messages]
        messages = '\n\n'.join(messages)
        return messages

    def clear(self):
        """Clear chat history and reduce messages to just the initial model
        role message."""
        self.messages = [{"role": "system", "content": self.MODEL_ROLE}]

    @staticmethod
    async def call_api(url, headers, request_json):
        """Make an asyncronous OpenAI API call.

        Parameters
        ----------
        url : str
            OpenAI API url, typically either:
                https://api.openai.com/v1/embeddings
                https://api.openai.com/v1/chat/completions
        headers : dict
            OpenAI API headers, typically:
                {"Content-Type": "application/json",
                 "Authorization": f"Bearer {openai.api_key}"}
        request_json : dict
            API data input, typically looks like this for chat completion:
                {"model": "gpt-3.5-turbo",
                 "messages": [{"role": "system", "content": "You do this..."},
                              {"role": "user", "content": "Do this: {}"}],
                 "temperature": 0.0}

        Returns
        -------
        out : dict
            API response in json format
        """

        out = None
        kwargs = dict(url=url, headers=headers, json=request_json)
        try:
            async with aiohttp.ClientSession() as session:
                async with session.post(**kwargs) as response:
                    out = await response.json()

        except Exception as e:
            logger.debug(f'Error in OpenAI API call from '
                         f'`aiohttp.ClientSession().post(**kwargs)` with '
                         f'kwargs: {kwargs}')
            logger.exception('Error in OpenAI API call! Turn on debug logging '
                             'to see full query that caused error.')
            out = {'error': str(e)}

        return out

    async def call_api_async(self, url, headers, all_request_jsons,
                             ignore_error=None, rate_limit=40e3):
        """Use GPT to clean raw pdf text in parallel calls to the OpenAI API.

        NOTE: you need to call this using the await command in ipython or
        jupyter, e.g.: `out = await PDFtoTXT.clean_txt_async()`

        Parameters
        ----------
        url : str
            OpenAI API url, typically either:
                https://api.openai.com/v1/embeddings
                https://api.openai.com/v1/chat/completions
        headers : dict
            OpenAI API headers, typically:
                {"Content-Type": "application/json",
                 "Authorization": f"Bearer {openai.api_key}"}
        all_request_jsons : list
            List of API data input, one entry typically looks like this for
            chat completion:
                {"model": "gpt-3.5-turbo",
                 "messages": [{"role": "system", "content": "You do this..."},
                              {"role": "user", "content": "Do this: {}"}],
                 "temperature": 0.0}
        ignore_error : None | callable
            Optional callable to parse API error string. If the callable
            returns True, the error will be ignored, the API call will not be
            tried again, and the output will be an empty string.
        rate_limit : float
            OpenAI API rate limit (tokens / minute). Note that the
            gpt-3.5-turbo limit is 90k as of 4/2023, but we're using a large
            factor of safety (~1/2) because we can only count the tokens on the
            input side and assume the output is about the same count.

        Returns
        -------
        out : list
            List of API outputs where each list entry is a GPT answer from the
            corresponding message in the all_request_jsons input.
        """
        self.api_queue = ApiQueue(url, headers, all_request_jsons,
                                  ignore_error=ignore_error,
                                  rate_limit=rate_limit)
        out = await self.api_queue.run()
        return out

    def chat(self, query, temperature=0):
        """Have a continuous chat with the LLM including context from previous
        chat() calls stored as attributes in this class.

        Parameters
        ----------
        query : str
            Question to ask ChatGPT
        temperature : float
            GPT model temperature, a measure of response entropy from 0 to 1. 0
            is more reliable and nearly deterministic; 1 will give the model
            more creative freedom and may not return as factual of results.

        Returns
        -------
        response : str
            Model response
        """

        self.messages.append({"role": "user", "content": query})

        kwargs = dict(model=self.model,
                      messages=self.messages,
                      temperature=temperature,
                      stream=False)

        response = self._client.chat.completions.create(**kwargs)
        response = response.choices[0].message.content
        self.messages.append({'role': 'assistant', 'content': response})

        return response

    def generic_query(self, query, model_role=None, temperature=0):
        """Ask a generic single query without conversation

        Parameters
        ----------
        query : str
            Question to ask ChatGPT
        model_role : str | None
            Role for the model to take, e.g.: "You are a research assistant".
            This defaults to self.MODEL_ROLE
        temperature : float
            GPT model temperature, a measure of response entropy from 0 to 1. 0
            is more reliable and nearly deterministic; 1 will give the model
            more creative freedom and may not return as factual of results.

        Returns
        -------
        response : str
            Model response
        """

        model_role = model_role or self.MODEL_ROLE
        messages = [{"role": "system", "content": model_role},
                    {"role": "user", "content": query}]
        kwargs = dict(model=self.model,
                      messages=messages,
                      temperature=temperature,
                      stream=False)

        response = self._client.chat.completions.create(**kwargs)
        response = response.choices[0].message.content
        return response

    async def generic_async_query(self, queries, model_role=None,
                                  temperature=0, ignore_error=None,
                                  rate_limit=40e3):
        """Run a number of generic single queries asynchronously
        (not conversational)

        NOTE: you need to call this using the await command in ipython or
        jupyter, e.g.: `out = await Summary.run_async()`

        Parameters
        ----------
        query : list
            Questions to ask ChatGPT (list of strings)
        model_role : str | None
            Role for the model to take, e.g.: "You are a research assistant".
            This defaults to self.MODEL_ROLE
        temperature : float
            GPT model temperature, a measure of response entropy from 0 to 1. 0
            is more reliable and nearly deterministic; 1 will give the model
            more creative freedom and may not return as factual of results.
        ignore_error : None | callable
            Optional callable to parse API error string. If the callable
            returns True, the error will be ignored, the API call will not be
            tried again, and the output will be an empty string.
        rate_limit : float
            OpenAI API rate limit (tokens / minute). Note that the
            gpt-3.5-turbo limit is 90k as of 4/2023, but we're using a large
            factor of safety (~1/2) because we can only count the tokens on the
            input side and assume the output is about the same count.

        Returns
        -------
        response : list
            Model responses with same length as query input.
        """

        model_role = model_role or self.MODEL_ROLE
        all_request_jsons = []
        for msg in queries:
            msg = [{'role': 'system', 'content': self.MODEL_ROLE},
                   {'role': 'user', 'content': msg}]
            req = {"model": self.model, "messages": msg,
                   "temperature": temperature}
            all_request_jsons.append(req)

        self.api_queue = ApiQueue(self.URL, self.HEADERS, all_request_jsons,
                                  ignore_error=ignore_error,
                                  rate_limit=rate_limit)
        out = await self.api_queue.run()

        for i, response in enumerate(out):
            choice = response.get('choices', [{'message': {'content': ''}}])[0]
            message = choice.get('message', {'content': ''})
            content = message.get('content', '')
            if not any(content):
                logger.error(f'Received no output for query {i + 1}!')
            else:
                out[i] = content

        return out

    @classmethod
    def get_embedding(cls, text):
        """Get the 1D array (list) embedding of a text string.

        Parameters
        ----------
        text : str
            Text to embed

        Returns
        -------
        embedding : list
            List of float that represents the numerical embedding of the text
        """
        kwargs = dict(url=cls.EMBEDDING_URL,
                      headers=cls.HEADERS,
                      json={'model': cls.EMBEDDING_MODEL,
                            'input': text})

        out = requests.post(**kwargs)
        embedding = out.json()

        try:
            embedding = embedding["data"][0]["embedding"]
        except Exception as exc:
            msg = ('Embedding request failed: {} {}'
                   .format(out.reason, embedding))
            logger.error(msg)
            raise RuntimeError(msg) from exc

        return embedding

    @classmethod
    def count_tokens(cls, text, model, fallback_model='gpt-4'):
        """Return the number of tokens in a string.

        Parameters
        ----------
        text : str
            Text string to get number of tokens for
        model : str
            specification of OpenAI model to use (e.g., "gpt-3.5-turbo")
        fallback_model : str, default='gpt-4'
            Model to be used for tokenizer if input model can't be found
            in :obj:`TOKENIZER_ALIASES` and doesn't have any easily
            noticeable patterns.

        Returns
        -------
        n : int
            Number of tokens in text
        """

        if model in cls.TOKENIZER_ALIASES:
            token_model = cls.TOKENIZER_ALIASES[model]
        else:
            token_model = fallback_model
            for pattern in cls.TOKENIZER_PATTERNS:
                if pattern in model:
                    token_model = pattern
                    break

        encoding = tiktoken.encoding_for_model(token_model)

        return len(encoding.encode(text))


class ApiQueue:
    """Class to manage the parallel API queue and submission"""

    def __init__(self, url, headers, request_jsons, ignore_error=None,
                 rate_limit=40e3, max_retries=10):
        """
        Parameters
        ----------
        url : str
            OpenAI API url, typically either:
                https://api.openai.com/v1/embeddings
                https://api.openai.com/v1/chat/completions
        headers : dict
            OpenAI API headers, typically:
                {"Content-Type": "application/json",
                 "Authorization": f"Bearer {openai.api_key}"}
        all_request_jsons : list
            List of API data input, one entry typically looks like this for
            chat completion:
                {"model": "gpt-3.5-turbo",
                 "messages": [{"role": "system", "content": "You do this..."},
                              {"role": "user", "content": "Do this: {}"}],
                 "temperature": 0.0}
        ignore_error : None | callable
            Optional callable to parse API error string. If the callable
            returns True, the error will be ignored, the API call will not be
            tried again, and the output will be an empty string.
        rate_limit : float
            OpenAI API rate limit (tokens / minute). Note that the
            gpt-3.5-turbo limit is 90k as of 4/2023, but we're using a large
            factor of safety (~1/2) because we can only count the tokens on the
            input side and assume the output is about the same count.
        max_retries : int
            Number of times to retry an API call with an error response before
            raising an error.
        """

        self.url = url
        self.headers = headers
        self.request_jsons = request_jsons
        self.ignore_error = ignore_error
        self.rate_limit = rate_limit
        self.max_retries = max_retries
        self.api_jobs = None
        self.todo = None
        self.out = None
        self.errors = None
        self.tries = None
        self._retry = False
        self._tsub = 0
        self._reset()
        self.job_names = [f'job_{str(ijob).zfill(4)}'
                          for ijob in range(len(request_jsons))]

    def _reset(self):
        self.api_jobs = {}
        self.todo = [True] * len(self)
        self.out = [None] * len(self)
        self.errors = [None] * len(self)
        self.tries = np.zeros(len(self), dtype=int)
        self._retry = False
        self._tsub = 0

    def __len__(self):
        """Number of API calls to submit"""
        return len(self.request_jsons)

    @property
    def waiting_on(self):
        """Get a list of async jobs that are being waited on."""
        return [job for ijob, job in self.api_jobs.items() if self.todo[ijob]]

    def submit_jobs(self):
        """Submit a subset jobs asynchronously and hold jobs in the `api_jobs`
        attribute. Break when the `rate_limit` is exceeded."""

        token_count = 0
        t_elap = (time.time() - self._tsub) / 60
        avail_tokens = self.rate_limit * t_elap
        avail_tokens = min(self.rate_limit, avail_tokens)

        for ijob, itodo in enumerate(self.todo):
            if (ijob not in self.api_jobs
                    and itodo
                    and token_count < avail_tokens):
                request = self.request_jsons[ijob]
                model = request['model']
                tokens = ApiBase.count_tokens(str(request), model)

                if tokens > self.rate_limit:
                    msg = ('Job index #{} with has {} tokens which '
                           'is greater than the rate limit of {}!'
                           .format(ijob, tokens, self.rate_limit))
                    logger.error(msg)
                    raise RuntimeError(msg)

                elif tokens < avail_tokens:
                    token_count += tokens
                    task = asyncio.create_task(ApiBase.call_api(self.url,
                                                                self.headers,
                                                                request),
                                               name=self.job_names[ijob])
                    self.api_jobs[ijob] = task
                    self.tries[ijob] += 1
                    self._tsub = time.time()

                    logger.debug('Submitted "{}" ({} out of {}). '
                                 'Token count: {} '
                                 '(rate limit is {}). '
                                 'Attempts: {}'
                                 .format(self.job_names[ijob],
                                         ijob + 1, len(self), token_count,
                                         self.rate_limit,
                                         self.tries[ijob]))

            elif token_count >= avail_tokens:
                token_count = 0
                break

    async def collect_jobs(self):
        """Collect asyncronous API calls and API outputs. Store outputs in the
        `out` attribute."""

        if not any(self.waiting_on):
            return

        complete, _ = await asyncio.wait(self.waiting_on,
                                         return_when=asyncio.FIRST_COMPLETED)

        for job in complete:
            job_name = job.get_name()
            ijob = self.job_names.index(job_name)
            task_out = job.result()

            if 'error' in task_out:
                msg = ('Received API error for task #{0} '
                       '(see `ApiQueue.errors[{1}]` and '
                       '`ApiQueue.request_jsons[{1}]` for more details). '
                       'Error message: {2}'.format(ijob + 1, ijob, task_out))
                self.errors[ijob] = 'Error: {}'.format(task_out)

                if (self.ignore_error is not None
                        and self.ignore_error(str(task_out))):
                    msg += ' Ignoring error and moving on.'
                    dummy = {'choices': [{'message': {'content': ''}}]}
                    self.out[ijob] = dummy
                    self.todo[ijob] = False
                else:
                    del self.api_jobs[ijob]
                    msg += ' Retrying query.'
                    self._retry = True
                logger.error(msg)

            else:
                self.out[ijob] = task_out
                self.todo[ijob] = False

        n_complete = len(self) - sum(self.todo)
        logger.debug('Finished {} API calls, {} left'
                     .format(n_complete, sum(self.todo)))

    async def run(self):
        """Run all asyncronous API calls.

        Returns
        -------
        out : list
            List of API call outputs with same ordering as `request_jsons`
            input.
        """

        self._reset()
        logger.debug('Submitting async API calls...')

        i = 0
        while any(self.todo):
            i += 1
            self._retry = False
            self.submit_jobs()
            await self.collect_jobs()

            if any(self.tries > self.max_retries):
                msg = (f'Hit {self.max_retries} retries on API queries. '
                       'Stopping. See `ApiQueue.errors` for more '
                       'details on error response')
                logger.error(msg)
                raise RuntimeError(msg)
            elif self._retry:
                time.sleep(10)
            elif i > 1e4:
                raise RuntimeError('Hit 1e4 iterations. What are you doing?')
            elif any(self.todo):
                time.sleep(5)

        return self.out



# -*- coding: utf-8 -*-
"""ELM parsing utilities."""
import io
import re
import logging
from warnings import warn

import html2text
import numpy as np
import pandas as pd


logger = logging.getLogger(__name__)


def is_multi_col(text, separator="    ", threshold_ratio=0.35):
    """Does the text look like it has multiple vertical text columns?

    Parameters
    ----------
    text : str
        Input text, which may or may not contain multiple vertical
        columns.
    separator : str
        Heuristic split string to look for spaces between columns
    threshold_ratio : float
        Portion of lines containing the separator at which point
        the text should be classified as multi-column.

    Returns
    -------
    out : bool
        True if more than one vertical text column
    """
    lines = text.split("\n")
    total_lines = len(lines)

    gap_lines = [line for line in lines if separator in line.strip()]
    cols = len(gap_lines)

    ratio = cols / total_lines

    return ratio >= threshold_ratio


def remove_blank_pages(pages):
    """Remove any blank pages from the iterable.

    Parameters
    ----------
    pages : iterable
        Iterable of string objects. Objects in this iterable that do not
        contain any text will be removed.

    Returns
    -------
    list
        List of strings with content, or empty list.
    """
    return [page for page in pages if any(page.strip())]


def html_to_text(html, ignore_links=True):
    """Call to `HTML2Text` class with basic args.

    Parameters
    ----------
    html : str
        HTML text extracted from the web.
    ignore_links : bool, optional
        Option to ignore links in HTML when parsing.
        By default, ``True``.

    Returns
    -------
    str
        Text extracted from the input HTML.
    """
    h = html2text.HTML2Text()
    h.ignore_links = ignore_links
    h.ignore_images = True
    h.bypass_tables = True
    return h.handle(html)


def format_html_tables(text, **kwargs):
    """Format tables within HTML text into pretty markdown.

    Note that if pandas does not detect enough tables in the text to
    match the "<table>" tags, no replacement is performed at all.

    Parameters
    ----------
    text : str
        HTML text, possible containing tables enclosed by the
        "<table>" tag.
    **kwargs
        Keyword-arguments to pass to ``pandas.DataFrame.to_markdown``
        function. Must not contain the `"headers"` keyword (this is
        supplied internally).

    Returns
    -------
    str
        Text with HTML tables (if any) converted to markdown.
    """
    matches = _find_html_table_matches(text)
    if not matches:
        return text

    dfs = _find_dfs(text)
    if len(matches) != len(dfs):
        logger.error(
            "Found incompatible number of HTML (%d) and parsed (%d) tables! "
            "No replacement performed.",
            len(matches),
            len(dfs),
        )
        return text

    return _replace_tables_in_text(text, matches, dfs, **kwargs)


def _find_html_table_matches(text):
    """Find HTML table matches in the text"""
    return re.findall(r"<table>[\s\S]*?</table>", text)


def _find_dfs(text):
    """Load HTML tables from text into DataFrames"""
    return pd.read_html(io.StringIO(text))


def _replace_tables_in_text(text, matches, dfs, **kwargs):
    """Replace all items in the 'matches' input with MD tables"""
    for table_str, df in zip(matches, dfs):
        new_table_str = df.to_markdown(headers=df.columns, **kwargs)
        text = text.replace(table_str, new_table_str)
    return text


def clean_headers(
    pages,
    char_thresh=0.6,
    page_thresh=0.8,
    split_on="\n",
    iheaders=(0, 1, -2, -1),
):
    """Clean headers/footers that are duplicated across pages of a document.

    Note that this function will update the items within the `pages`
    input.

    Parameters
    ----------
    pages : list
        List of pages (as str) from document.
    char_thresh : float
        Fraction of characters in a given header that are similar
        between pages to be considered for removal
    page_thresh : float
        Fraction of pages that share the header to be considered for
        removal
    split_on : str
        Chars to split lines of a page on
    iheaders : list | tuple
        Integer indices to look for headers after splitting a page into
        lines based on split_on. This needs to go from the start of the
        page to the end.

    Returns
    -------
    out : str
        Clean text with all pages joined
    """
    logger.info("Cleaning headers")
    headers = _get_nominal_headers(pages, split_on, iheaders)
    tests = np.zeros((len(pages), len(headers)))

    for ip, page in enumerate(pages):
        for ih, header in zip(iheaders, headers):
            pheader = ""
            try:
                pheader = page.split(split_on)[ih]
            except IndexError:
                pass

            harr = header.replace(" ", "")
            parr = pheader.replace(" ", "")

            harr = harr.ljust(len(parr))
            parr = parr.ljust(len(harr))

            harr = np.array([*harr])
            parr = np.array([*parr])
            assert len(harr) == len(parr)

            test = harr == parr
            if len(test) == 0:
                test = 1.0
            else:
                test = test.sum() / len(test)

            tests[ip, ih] = test

    logger.debug("Header tests (page, iheader): \n{}".format(tests))
    tests = (tests > char_thresh).sum(axis=0) / len(pages)
    tests = tests > page_thresh
    logger.debug("Header tests (iheader,): \n{}".format(tests))

    header_inds_to_remove = {
        ind for is_header, ind in zip(tests, iheaders) if is_header
    }
    if not header_inds_to_remove:
        return pages

    for ip, page in enumerate(pages):
        page = page.split(split_on)
        if len(iheaders) >= len(page):
            continue
        pages[ip] = split_on.join(
            [
                line
                for line_ind, line in enumerate(page)
                if line_ind not in header_inds_to_remove
                and line_ind - len(page) not in header_inds_to_remove
            ]
        )

    return pages


def _get_nominal_headers(pages, split_on, iheaders):
    """Get nominal headers from a standard page.

    This function aims for a "typical" page that is likely to have a
    normal header, not the first or last.

    Parameters
    ----------
    pages : list
        List of pages (as str) from document.
    split_on : str
        Chars to split lines of a page on
    iheaders : list | tuple
        Integer indices to look for headers after splitting a page into
        lines based on split_on. This needs to go from the start of the
        page to the end.

    Returns
    -------
    headers : list
        List of headers where each entry is a string header
    """

    headers = [None] * len(iheaders)
    page_lens = np.array([len(p) for p in pages])
    median_len = np.median(page_lens)
    ipage = np.argmin(np.abs(page_lens - median_len))
    page = pages[ipage]
    for i, ih in enumerate(iheaders):
        try:
            header = page.split(split_on)[ih]
        except IndexError:
            header = ""
        headers[i] = header

    return headers


def combine_pages(pages):
    """Combine pages of GPT cleaned text into a single string.

    Parameters
    ----------
    pages : list
        List of pages (as str) from document.

    Returns
    -------
    full : str
        Single multi-page string
    """
    return "\n".join(pages).replace("\n•", "-").replace("•", "-")


def replace_common_pdf_conversion_chars(text):
    """Re-format text to remove common pdf-converter chars.

    Chars affected include ``\\r\\n``, ``\\r`` and ``\\x0c``.

    Parameters
    ----------
    text : str
        Input text (presumably from pdf parser).

    Returns
    -------
    str
        Cleaned text.
    """
    return text.replace("\r\n", "\n").replace("\x0c", "").replace("\r", "\n")


def replace_multi_dot_lines(text):
    """Replace instances of three or more dots (.....) with just "..."

    Parameters
    ----------
    text : str
        Text possibly containing many repeated dots.

    Returns
    -------
    str
        Cleaned text with only three dots max in a row.
    """
    return re.sub(r"[.]{3,}", "...", text)


def replace_excessive_newlines(text):
    """Replace instances of three or more newlines with ``\\n\\n``

    Parameters
    ----------
    text : str
        Text possibly containing many repeated newline characters.

    Returns
    -------
    str
        Cleaned text with only a maximum of two newlines in a row.
    """
    return re.sub(r"[\n]{3,}", "\n\n", text)


def remove_empty_lines_or_page_footers(text):
    """Replace empty lines (potentially with page numbers only) as newlines

    Parameters
    ----------
    text : str
        Text possibly containing empty lines and/or lines with only page
        numbers.

    Returns
    -------
    str
        Cleaned text with no empty lines.
    """
    return re.sub(r"[\n\r]+(?:\s*?\d*?\s*)[\n\r]+", "\n", text)


def read_pdf(pdf_bytes, verbose=True):
    """Read PDF contents from bytes.

    This method will automatically try to detect multi-column format
    and load the text without a physical layout in that case.

    Parameters
    ----------
    pdf_bytes : bytes
        Bytes corresponding to a PDF file.
    verbose : bool, optional
        Option to log errors during parsing. By default, ``True``.

    Returns
    -------
    iterable
        Iterable containing pages of the PDF document. This iterable
        may be empty if there was an error reading the PDF file.
    """
    import pdftotext

    logger.trace("Attempting to load %d bytes as pdf", len(pdf_bytes))
    try:
        pages = _load_pdf_possibly_multi_col(pdf_bytes)
    except pdftotext.Error as e:
        if verbose:
            logger.error("Failed to decode PDF content!")
            logger.exception(e)
        pages = []

    logger.trace("`read_pdf` loaded %d bytes (%d pages)", len(pdf_bytes),
                 len(pages))
    return pages


def _load_pdf_possibly_multi_col(pdf_bytes):
    """Load PDF, which may be multi-column"""
    import pdftotext

    pdf_bytes = io.BytesIO(pdf_bytes)
    pages = pdftotext.PDF(pdf_bytes, physical=True)
    if is_multi_col(combine_pages(pages)):
        pages = pdftotext.PDF(pdf_bytes, physical=False)
    return pages


def read_pdf_ocr(  # pragma: no cover
    pdf_bytes,
    image_to_string_kwargs=None,
    convert_from_bytes_kwargs=None,
    verbose=True
):
    """Read PDF contents from bytes using Optical Character recognition (OCR).

    This method attempt to read the PDF document using OCR. This is one
    of the only ways to parse a scanned PDF document. To use this
    function, you will need to install the `pytesseract` and `pdf2image`
    Modules. Installation guides here:

        - `pytesseract`:
         https://github.com/madmaze/pytesseract?tab=readme-ov-file#installation
        - `pdf2image`:
         https://github.com/Belval/pdf2image?tab=readme-ov-file#how-to-install

    Windows users may also need to apply the fix described in this
    answer before they can use pytesseract: http://tinyurl.com/v9xr4vrj

    Parameters
    ----------
    pdf_bytes : bytes
        Bytes corresponding to a PDF file.
    image_to_string_kwargs : dictionary, optional
        Optional dictionary of keyword-value pairs to pass as arguments
        to the :func:`pytesseract.image_to_string` function.
        By default, ``None``.
    convert_from_bytes_kwargs : dictionary, optional
        Optional dictionary of keyword-value pairs to pass as arguments
        to the :func:`pdf2image.convert_from_bytes` function.
        By default, ``None``.
    verbose : bool, optional
        Option to log errors during parsing. By default, ``True``.

    Returns
    -------
    iterable
        Iterable containing pages of the PDF document. This iterable
        may be empty if there was an error reading the PDF file.
    """
    try:
        pages = _load_pdf_with_pytesseract(
            pdf_bytes,
            image_to_string_kwargs=image_to_string_kwargs,
            convert_from_bytes_kwargs=convert_from_bytes_kwargs
        )
    except Exception as e:
        if verbose:
            logger.error("Failed to decode PDF content!")
            logger.exception(e)
        pages = []

    return pages


def _load_pdf_with_pytesseract(  # pragma: no cover
    pdf_bytes, image_to_string_kwargs=None, convert_from_bytes_kwargs=None
):
    """Load PDF bytes using Optical Character recognition (OCR)"""

    try:
        import pytesseract
    except ImportError:
        msg = (
            "Module `pytesseract` not found. Please follow these instructions "
            "to install: https://github.com/madmaze/pytesseract?"
            "tab=readme-ov-file#installation"
        )
        logger.warning(msg)
        warn(msg)
        return []

    try:
        from pdf2image import convert_from_bytes
    except ImportError:
        msg = (
            "Module `pdf2image` not found. Please follow these instructions "
            "to install: https://github.com/Belval/pdf2image?"
            "tab=readme-ov-file#how-to-install"
        )
        logger.warning(msg)
        warn(msg)
        return []

    logger.debug(
        "Loading PDF with `tesseract_cmd` as %s",
        pytesseract.pytesseract.tesseract_cmd,
    )

    its = {"timeout": 60 * 5}
    its.update(image_to_string_kwargs or {})

    cfb = {"grayscale": True}
    cfb.update(convert_from_bytes_kwargs or {})
    return [
        str(pytesseract.image_to_string(page_data, **its).encode("utf-8"))
        for page_data in convert_from_bytes(bytes(pdf_bytes), **cfb)
    ]



# -*- coding: utf-8 -*-
"""ELM Ordinance queued logging.

This module implements queued logging, mostly following this blog:"
https://www.zopatista.com/python/2019/05/11/asyncio-logging/
"""
import asyncio
import logging
from pathlib import Path
from queue import SimpleQueue
from logging.handlers import QueueHandler, QueueListener


LOGGING_QUEUE = SimpleQueue()


class NoLocationFilter(logging.Filter):
    """Filter that catches all records without a location attribute."""

    def filter(self, record):
        """Filter logging record.

        Parameters
        ----------
        record : logging.LogRecord
            Log record containing the log message + default attributes.
            If the ``location`` attribute is missing or is a string in
            the form "Task-XX", the filter returns ``True`` (i.e. record
            is emitted).

        Returns
        -------
        bool
            If the record's ``location`` attribute is "missing".
        """
        record_location = getattr(record, "location", None)
        return record_location is None or "Task-" in record_location


class LocationFilter(logging.Filter):
    """Filter down to logs from a coroutine processing a specific location."""

    def __init__(self, location):
        """

        Parameters
        ----------
        location : str
            Location identifier. For example, ``"El Paso Colorado"``.
        """
        self.location = location

    def filter(self, record):
        """Filter logging record.

        Parameters
        ----------
        record : logging.LogRecord
            Log record containing the log message + default attributes.
            Must have a ``location`` attribute that is a string
            identifier, or this function will return ``False`` every
            time. The ``location`` identifier will be checked against
            the filter's location attribute to determine the output
            result.

        Returns
        -------
        bool
            If the record's ``location`` attribute matches the filter's
            ``location`` attribute.
        """
        record_location = getattr(record, "location", None)
        return record_location is not None and record_location == self.location


class LocalProcessQueueHandler(QueueHandler):
    """QueueHandler that works within a single process (locally)."""

    def emit(self, record):
        """Emit record with a location attribute equal to current asyncio task.

        Parameters
        ----------
        record : logging.LogRecord
            Log record containing the log message + default attributes.
            This record will get a ``location`` attribute dynamically
            added, with a value equal to the name of the current asyncio
            task (i.e. ``asyncio.current_task().get_name()``).
        """
        try:
            record.location = asyncio.current_task().get_name()
        except RuntimeError:
            record.location = None

        try:
            self.enqueue(record)
        except asyncio.CancelledError:
            raise
        except Exception:
            self.handleError(record)


class LogListener:
    """Class to listen to logging queue from coroutines and write to files."""

    def __init__(self, logger_names, level="INFO"):
        """

        Parameters
        ----------
        logger_names : iterable
            An iterable of string, where each string is a logger name.
            The logger corresponding to each of the names will be
            equipped with a logging queue handler.
        level : str, optional
            Log level to set for each logger. By default, ``"INFO"``.
        """
        self.logger_names = logger_names
        self.level = level
        self._listener = None
        self._queue_handler = LocalProcessQueueHandler(LOGGING_QUEUE)

    def _setup_listener(self):
        """Set up the queue listener"""
        if self._listener is not None:
            return
        self._listener = QueueListener(
            LOGGING_QUEUE, logging.NullHandler(), respect_handler_level=True
        )
        self._listener.handlers = list(self._listener.handlers)

    def _add_queue_handler_to_loggers(self):
        """Add a queue handler to each logger requested by user"""
        for logger_name in self.logger_names:
            logger = logging.getLogger(logger_name)
            logger.addHandler(self._queue_handler)
            logger.setLevel(self.level)

    def _remove_queue_handler_from_loggers(self):
        """Remove the queue handler from each logger requested by user"""
        for logger_name in self.logger_names:
            logging.getLogger(logger_name).removeHandler(self._queue_handler)

    def _remove_all_handlers_from_listener(self):
        """Remove all handlers still attached to listener."""
        if self._listener is None:
            return
        for handler in self._listener.handlers:
            handler.close()
            self._listener.handlers.remove(handler)

    def __enter__(self):
        self._setup_listener()
        self._add_queue_handler_to_loggers()
        self._listener.start()
        return self

    def __exit__(self, exc_type, exc, tb):
        self._listener.stop()
        self._remove_queue_handler_from_loggers()
        self._remove_all_handlers_from_listener()

    async def __aenter__(self):
        return self.__enter__()

    async def __aexit__(self, exc_type, exc, tb):
        self.__exit__(exc_type, exc, tb)

    def addHandler(self, handler):
        """Add a handler to the queue listener.

        Logs that are sent to the queue will be emitted to the handler.

        Parameters
        ----------
        handler : logging.Handler
            Log handler to parse log records.
        """
        if handler not in self._listener.handlers:
            self._listener.handlers.append(handler)

    def removeHandler(self, handler):
        """Remove a handler from the queue listener.

        Logs that are sent to the queue will no longer be emitted to the
        handler.

        Parameters
        ----------
        handler : logging.Handler
            Log handler to remove from queue listener.
        """
        if handler in self._listener.handlers:
            handler.close()
            self._listener.handlers.remove(handler)


class LocationFileLog:
    """Context manager to write logs for a location to a unique file."""

    def __init__(self, listener, log_dir, location, level="INFO"):
        """

        Parameters
        ----------
        listener : :class:`~elm.ords.utilities.queued_logging.LoggingListener`
            A listener instance. The file handler will be added to this
            listener.
        log_dir : path-like
            Path to output directory to contain log file.
        location : str
            Location identifier. For example, ``"El Paso Colorado"``.
            This string will become part of the file name, so it must
            contain only characters valid in a file name.
        level : str, optional
            Log level. By default, ``"INFO"``.
        """
        self.log_dir = Path(log_dir)
        self.location = location
        self.level = level
        self._handler = None
        self._listener = listener

    def _create_log_dir(self):
        """Create log output directory if it doesn't exist."""
        self.log_dir.mkdir(exist_ok=True, parents=True)

    def _setup_handler(self):
        """Setup the file handler for this location."""
        self._handler = logging.FileHandler(
            self.log_dir / f"{self.location}.log", encoding="utf-8"
        )
        self._handler.setLevel(self.level)
        self._handler.addFilter(LocationFilter(self.location))

    def _break_down_handler(self):
        """Tear down the file handler for this location."""
        if self._handler is None:
            return

        self._handler.close()
        self._handler = None

    def _add_handler_to_listener(self):
        """Add the file handler for this location to the queue listener."""
        if self._handler is None:
            raise ValueError("Must set up handler before listener!")

        self._listener.addHandler(self._handler)

    def _remove_handler_from_listener(self):
        """Remove the file handler for this location from the listener."""
        if self._handler is None:
            return

        self._listener.removeHandler(self._handler)

    def __enter__(self):
        self._create_log_dir()
        self._setup_handler()
        self._add_handler_to_listener()

    def __exit__(self, exc_type, exc, tb):
        self._remove_handler_from_listener()
        self._break_down_handler()

    async def __aenter__(self):
        self.__enter__()

    async def __aexit__(self, exc_type, exc, tb):
        self.__exit__(exc_type, exc, tb)



# -*- coding: utf-8 -*-
"""ELM Web Document class definitions"""
import re
from abc import ABC, abstractmethod
from copy import deepcopy
from functools import cached_property
import logging

import pandas as pd

from elm.utilities.parse import (
    combine_pages,
    clean_headers,
    html_to_text,
    remove_blank_pages,
    format_html_tables,
    read_pdf,
    read_pdf_ocr,
    replace_common_pdf_conversion_chars,
    replace_multi_dot_lines,
    remove_empty_lines_or_page_footers,
)


logger = logging.getLogger(__name__)


class BaseDocument(ABC):
    """Base ELM web document representation

    Purpose:
        Track document content and perform minor processing on it.
    Responsibilities:
        1. Store "raw" document text.
        2. Compute "cleaned" text, which combines pages, strips HTML,
           and formats tables.
        3. Track pages and other document metadata.
    Key Relationships:
        Created by :class:`~elm.web.file_loader.AsyncFileLoader` and
        used all over ordinance code.

    .. end desc
    """

    def __init__(self, pages, attrs=None):
        """

        Parameters
        ----------
        pages : iterable
            Iterable of strings, where each string is a page of a
            document.
        attrs : dict, optional
            Optional dict containing metadata for the document.
            By default, ``None``.
        """
        self.pages = remove_blank_pages(pages)
        self.attrs = attrs or {}

    def __repr__(self):
        header = (f"{self.__class__.__name__} with {len(self.pages):,} "
                  "pages\nAttrs:")
        if not self.attrs:
            return f"{header} None"

        attrs = {}
        for k, v in self.attrs.items():
            if isinstance(v, pd.DataFrame):
                v = f"DataFrame with {len(v):,} rows"
            attrs[k] = v

        indent = max(len(k) for k in attrs) + 2
        attrs = "\n".join([f"{k:>{indent}}:\t{v}"
                           for k, v in attrs.items()])
        return f"{header}\n{attrs}"

    @property
    def empty(self):
        """bool: ``True`` if the document contains no pages."""
        return not any(_non_empty_pages(self.text.split("\n")))

    @cached_property
    def raw_pages(self):
        """list: List of (a limited count of) raw pages"""
        if not self.pages:
            return []

        return self._raw_pages()

    @cached_property
    def text(self):
        """str: Cleaned text from document"""
        if not self.pages:
            return ""

        return self._cleaned_text()

    @abstractmethod
    def _raw_pages(self):
        """Get raw pages from document"""
        raise NotImplementedError(
            "This document does not implement a raw pages extraction function"
        )

    @abstractmethod
    def _cleaned_text(self):
        """Compute cleaned text from document"""
        raise NotImplementedError(
            "This document does not implement a pages cleaning function"
        )

    @property
    @abstractmethod
    def WRITE_KWARGS(self):
        """dict: Dict of kwargs to pass to `open` when writing this doc."""
        raise NotImplementedError

    @property
    @abstractmethod
    def FILE_EXTENSION(self):
        """str: Cleaned document file extension."""
        raise NotImplementedError


class PDFDocument(BaseDocument):
    """ELM web PDF document"""

    CLEAN_HEADER_KWARGS = {
        "char_thresh": 0.6,
        "page_thresh": 0.8,
        "split_on": "\n",
        "iheaders": [0, 1, 3, -3, -2, -1],
    }
    """Default :func:`~elm.utilities.parse.clean_headers` arguments"""
    WRITE_KWARGS = {"mode": "wb"}
    FILE_EXTENSION = "pdf"

    def __init__(
        self,
        pages,
        attrs=None,
        percent_raw_pages_to_keep=25,
        max_raw_pages=18,
        num_end_pages_to_keep=2,
        clean_header_kwargs=None,
    ):
        """

        Parameters
        ----------
        pages : iterable
            Iterable of strings, where each string is a page of a
            document.
        attrs : str, optional
            Optional dict containing metadata for the document.
            By default, ``None``.
        percent_raw_pages_to_keep : int, optional
            Percent of "raw" pages to keep. Useful for extracting info
            from headers/footers of a doc, which are normally stripped
            to form the "clean" text. By default, ``25``.
        max_raw_pages : int, optional
            The max number of raw pages to keep. The number of raw pages
            will never exceed the total of this value +
            `num_end_pages_to_keep`. By default, ``18``.
        num_end_pages_to_keep : int, optional
            Number of additional pages to keep from the end of the
            document. This can be useful to extract more meta info.
            The number of raw pages will never exceed the total of this
            value + `max_raw_pages`. By default, ``2``.
        clean_header_kwargs : dict, optional
            Optional dictionary of keyword-value pair arguments to pass
            to the :func:`~elm.utilities.parse.clean_headers`
            function. By default, ``None``.
        """
        super().__init__(pages, attrs=attrs)
        self.percent_raw_pages_to_keep = percent_raw_pages_to_keep
        self.max_raw_pages = min(len(self.pages), max_raw_pages)
        self.num_end_pages_to_keep = num_end_pages_to_keep
        self.clean_header_kwargs = deepcopy(self.CLEAN_HEADER_KWARGS)
        self.clean_header_kwargs.update(clean_header_kwargs or {})

    @cached_property
    def num_raw_pages_to_keep(self):
        """int: Number of raw pages to keep from PDF document"""
        num_to_keep = self.percent_raw_pages_to_keep / 100 * len(self.pages)
        return min(self.max_raw_pages, max(1, int(num_to_keep)))

    @cached_property
    def _last_page_index(self):
        """int: last page index (determines how many end pages to include)"""
        neg_num_extra_pages = self.num_raw_pages_to_keep - len(self.pages)
        neg_num_last_pages = max(
            -self.num_end_pages_to_keep, neg_num_extra_pages
        )
        return min(0, neg_num_last_pages)

    def _cleaned_text(self):
        """Compute cleaned text from document"""
        pages = clean_headers(deepcopy(self.pages), **self.clean_header_kwargs)
        text = combine_pages(pages)
        text = replace_common_pdf_conversion_chars(text)
        text = replace_multi_dot_lines(text)
        text = remove_empty_lines_or_page_footers(text)
        return text

    # pylint: disable=unnecessary-comprehension
    # fmt: off
    def _raw_pages(self):
        """Get raw pages from document"""
        raw_pages = [page for page in self.pages[:self.num_raw_pages_to_keep]]
        if self._last_page_index:
            raw_pages += [page for page in self.pages[self._last_page_index:]]
        return raw_pages

    @classmethod
    def from_file(cls, fp, **init_kwargs):
        """Initialize a PDFDocument object from a .pdf file on disk. This
        method will try to use pdftotext (a poppler utility) and then
        OCR with pytesseract.

        Parameters
        ----------
        fp : str
            filepath to .pdf on disk
        init_kwargs : dict
            Optional kwargs for PDFDocument Initialization

        Returns
        -------
        out : PDFDocument
            Initialized PDFDocument class from input fp
        """

        with open(fp, 'rb') as f:
            pages = read_pdf(f.read())

        pages = list(_non_empty_pages(pages))
        if pages:
            return cls(pages, **init_kwargs)

        # fallback to OCR with pytesseract if no pages have more than 10
        # chars. Typical scanned document only has weird ascii per page.
        with open(fp, 'rb') as f:
            pages = read_pdf_ocr(f.read())

        pages = list(_non_empty_pages(pages))
        if not any(pages):
            msg = f'Could not get text from pdf: {fp}'
            logger.error(msg)
            raise RuntimeError(msg)

        return cls(pages, **init_kwargs)


class HTMLDocument(BaseDocument):
    """ELM web HTML document"""

    HTML_TABLE_TO_MARKDOWN_KWARGS = {
        "floatfmt": ".5f",
        "index": True,
        "tablefmt": "psql",
    }
    """Default :func:`~elm.utilities.parse.format_html_tables` arguments"""
    WRITE_KWARGS = {"mode": "w", "encoding": "utf-8"}
    FILE_EXTENSION = "txt"

    def __init__(
        self,
        pages,
        attrs=None,
        html_table_to_markdown_kwargs=None,
        ignore_html_links=True,
        text_splitter=None,
    ):
        """

        Parameters
        ----------
        pages : iterable
            Iterable of strings, where each string is a page of a
            document.
        attrs : dict, optional
            Optional dict containing metadata for the document.
            By default, ``None``.
        html_table_to_markdown_kwargs : dict, optional
            Optional dictionary of keyword-value pair arguments to pass
            to the :func:`~elm.utilities.parse.format_html_tables`
            function. By default, ``None``.
        ignore_html_links : bool, optional
            Option to ignore link in HTML text during parsing.
            By default, ``True``.
        text_splitter : obj, optional
            Instance of an object that implements a `split_text` method.
            The method should take text as input (str) and return a list
            of text chunks. The raw pages will be passed through this
            splitter to create raw pages for this document. Langchain's
            text splitters should work for this input.
            By default, ``None``, which means the original pages input
            becomes the raw pages attribute.
        """
        super().__init__(pages, attrs=attrs)
        self.html_table_to_markdown_kwargs = deepcopy(
            self.HTML_TABLE_TO_MARKDOWN_KWARGS
        )
        self.html_table_to_markdown_kwargs.update(
            html_table_to_markdown_kwargs or {}
        )
        self.ignore_html_links = ignore_html_links
        self.text_splitter = text_splitter

    def _cleaned_text(self):
        """Compute cleaned text from document"""
        text = combine_pages(self.pages)
        text = html_to_text(text, self.ignore_html_links)
        text = format_html_tables(text, **self.html_table_to_markdown_kwargs)
        return text

    def _raw_pages(self):
        """Get raw pages from document"""
        if self.text_splitter is None:
            return self.pages
        return self.text_splitter.split_text("\n\n".join(self.pages))


def _non_empty_pages(pages):
    """Return all pages with more than 10 chars"""
    return filter(
        lambda page: re.search('[a-zA-Z]', page) and len(page) > 10, pages
    )



"""ELM validation utilities."""
import os


def validate_azure_api_params(azure_api_key=None, azure_version=None,
                              azure_endpoint=None):
    """Validate Azure OpenAI API parameters.

    Parameters
    ----------
    azure_api_key, azure_version, azure_endpoint : str, optional
        Azure OpenAI API key, version, and endpoint. By default,
        ``None``, which attempts to read these variables from the
        following environment variables:

            - AZURE_OPENAI_API_KEY
            - AZURE_OPENAI_VERSION
            - AZURE_OPENAI_ENDPOINT

        If any of these are still `None` after reading from the
        environment, an error is raised.

    Returns
    -------
    azure_api_key, azure_version, azure_endpoint : str
        API key, version, and endpoint that can be used to initialize
        an Azure OpenAI service.
    """
    azure_api_key = azure_api_key or os.environ.get("AZURE_OPENAI_API_KEY")
    azure_version = azure_version or os.environ.get("AZURE_OPENAI_VERSION")
    azure_endpoint = azure_endpoint or os.environ.get("AZURE_OPENAI_ENDPOINT")
    assert azure_api_key is not None, "Must set AZURE_OPENAI_API_KEY!"
    assert azure_version is not None, "Must set AZURE_OPENAI_VERSION!"
    assert azure_endpoint is not None, "Must set AZURE_OPENAI_ENDPOINT!"
    return azure_api_key, azure_version, azure_endpoint



# -*- coding: utf-8 -*-
"""ELM Web Scraping utilities."""
import uuid
import hashlib
import logging
import asyncio
from pathlib import Path
from copy import deepcopy
from random import randint, choice
from contextlib import asynccontextmanager

import httpx
from slugify import slugify
from fake_useragent import UserAgent
from playwright_stealth import stealth_async
from scrapling.engines import PlaywrightEngine

from elm.web.document import PDFDocument


logger = logging.getLogger(__name__)
DEFAULT_HEADERS = {
    "Accept": (
        "text/html,application/xhtml+xml,application/xml;"
        "q=0.9,image/webp,*/*;q=0.8"
    ),
    "Accept-Language": "en-US,en;q=0.5",
    "Referer": "https://www.google.com/",
    "DNT": "1",
    "Connection": "keep-alive",
    "Upgrade-Insecure-Requests": "1",
}
"""Default HTML header template"""
_BT_RENAME = {"chromium": "Chrome", "firefox": "Firefox",
              "webkit": "Safari"}
# block pages by resource type. e.g. image, stylesheet
BLOCK_RESOURCE_TYPES = [
    "beacon",
    "csp_report",
    "font",
    "image",
    "imageset",
    "media",
    "object",
    "texttrack",
    'websocket',
    #  can block stylsheets and scripts, though it's not recommended:
    # 'stylesheet',
    # 'script',
    # 'xhr',
]
# block popular 3rd party resources like tracking and advertisements.
BLOCK_RESOURCE_NAMES = [
    "adzerk",
    "analytics",
    "cdn.api.twitter",
    "doubleclick",
    "exelator",
    "facebook",
    "fontawesome",
    "google",
    "google-analytics",
    "googletagmanager",
    "lit.connatix",  # <- not sure about this one
]


async def get_redirected_url(url, **kwargs):
    """Get the final URL after following redirects.

    Parameters
    ----------
    url : str
        URL to check for redirects.
    **kwargs
        Keyword-value arguments to pass to the
        :class:`httpx.AsyncClient` client.

    Returns
    -------
    str
        The final URL after following redirects, or the original URL if
        no redirects are found or an error occurs.
    """
    kwargs["follow_redirects"] = True
    try:
        async with httpx.AsyncClient(**kwargs) as client:
            response = await client.head(url)
            return str(response.url)
    except httpx.RequestError as e:
        return url


def clean_search_query(query):
    """Check if the first character is a digit and remove it if so.

    Some search tools (e.g., Google) will fail to return results if the
    query has a leading digit: 1. "LangCh..."

    This function will take all the text after the first double quote
    (") if a digit is detected at the beginning of the string.

    Parameters
    ----------
    query : str
        Input query that may or may not contain a leading digit.

    Returns
    -------
    str
        Cleaned query.
    """
    query = query.strip()
    if len(query) < 1:
        return query

    if not query[0].isdigit():
        return query.strip()

    if (first_quote_pos := query[:-1].find('"')) == -1:
        return query.strip()

    last_ind = -1 if query.endswith('"') else None

    # fmt: off
    return query[first_quote_pos + 1:last_ind].strip()


def compute_fn_from_url(url, make_unique=False):
    """Compute a unique file name from URL string.

    File name will always be 128 characters or less, unless the
    `make_unique` argument is set to true. In that case, the max
    length is 164 (a UUID is tagged onto the filename).

    Parameters
    ----------
    url : str
        Input URL to convert into filename.
    make_unique : bool, optional
        Option to add a UUID at the end of the file name to make it
        unique. By default, ``False``.

    Returns
    -------
    str
        Valid filename representation of the URL.
    """
    url = url.replace("https", "").replace("http", "").replace("www", "")
    url = slugify(url)
    url = url.replace("-", "").replace("_", "")

    url = _shorten_using_sha(url)

    if make_unique:
        url = f"{url}{uuid.uuid4()}".replace("-", "")

    return url


def _shorten_using_sha(fn):
    """Reduces FN to 128 characters"""
    if len(fn) <= 128:
        return fn

    out = hashlib.sha256(bytes(fn[64:], encoding="utf-8")).hexdigest()
    return f"{fn[:64]}{out}"


def write_url_doc_to_file(doc, file_content, out_dir, make_name_unique=False):
    """Write a file pulled from URL to disk.

    Parameters
    ----------
    doc : elm.web.document.Document
        Document containing meta information about the file. Must have a
        "source" key in the `metadata` dict containing the URL, which
        will be converted to a file name using
        :func:`compute_fn_from_url`.
    file_content : str | bytes
        File content, typically string text for HTML files and bytes
        for PDF file.
    out_dir : path-like
        Path to directory where file should be stored.
    make_name_unique : bool, optional
        Option to make file name unique by adding a UUID at the end of
        the file name. By default, ``False``.

    Returns
    -------
    Path
        Path to output file.
    """
    out_fn = compute_fn_from_url(
        url=doc.attrs["source"], make_unique=make_name_unique
    )
    out_fp = Path(out_dir) / f"{out_fn}.{doc.FILE_EXTENSION}"
    with open(out_fp, **doc.WRITE_KWARGS) as fh:
        fh.write(file_content)
    return out_fp


@asynccontextmanager
async def pw_page(browser, intercept_routes=False, stealth_config=None,
                  ignore_https_errors=False, timeout=30000,
                  use_scrapling_stealth=False):
    """Create new page from playwright browser context

    Parameters
    ----------
    browser : :class:`playwright.Browser`
        A playwright browser instance.
    intercept_routes : bool, default=False
        Option to intercept all requests and abort blocked ones.
        Be default, ``False``.
    stealth_config : :class:`playwright_stealth.Stealth`, optional
        Optional tf-playwright-stealth StealthConfig configuration
        object instance. By default, ``None``, which uses all the
        default stealth options.
    ignore_https_errors : bool, default=False
        Option to ignore https errors (i.e. SSL cert errors). This is
        not generally safe to do - you are susceptible to MITM attacks.
        However, if you are doing a simple scrape without providing
        any sensitive information (which you probably shouldn't be doing
        programmatically anyways), then it's probably ok to ignore these
        errors. By default, ``False``.
    timeout : int, default=30,000
        Default navigation and page load timeout (in milliseconds) to
        assign to page instance. By default, ``30_000``.
    use_scrapling_stealth : bool, default=False
        Option to use scrapling stealth scripts instead of
        tf-playwright-stealth. If set to ``True``, the `stealth_config`
        argument will be ignored. By default, ``False``.

    Yields
    ------
    :class:`playwright.Page`
        A new page that can be used for visiting websites.
    """
    browser_type = _BT_RENAME.get(browser.browser_type.name, "random")
    ck = PWKwargs.context_kwargs(browser_type=browser_type,
                                 ignore_https_errors=ignore_https_errors)

    context = await browser.new_context(**ck)

    try:
        logger.trace("Loading browser page")
        page = await context.new_page()
        page.set_default_navigation_timeout(timeout)
        page.set_default_timeout(timeout)

        await page.set_extra_http_headers(DEFAULT_HEADERS)
        if use_scrapling_stealth:
            logger.trace("Using scrapling stealth scripts")
            for script in PWKwargs.stealth_scripts():
                await page.add_init_script(path=script)
        else:
            logger.trace("Using tf-playwright-stealth stealth scripts")
            await stealth_async(page, stealth_config)

        if intercept_routes:
            logger.trace("Intercepting requests and aborting blocked ones")
            await page.route("**/*", _intercept_route)
        yield page
    finally:
        await context.close()


async def _intercept_route(route):  # pragma: no cover
    """intercept all requests and abort blocked ones

    Source: https://scrapfly.io/blog/how-to-block-resources-in-playwright/
    """
    if route.request.resource_type in BLOCK_RESOURCE_TYPES:
        return await route.abort()

    if any(key in route.request.url for key in BLOCK_RESOURCE_NAMES):
        return await route.abort()

    return await route.continue_()


async def filter_documents(
    documents, validation_coroutine, task_name=None, **kwargs
):
    """Filter documents by applying a filter function to each.

    Parameters
    ----------
    documents : iter of :class:`elm.web.document.BaseDocument`
        Iterable of documents to filter.
    validation_coroutine : coroutine
        A coroutine that returns ``False`` if the document should be
        discarded and ``True`` otherwise. This function should take a
        single :class:`elm.web.document.BaseDocument` instance as the
        first argument. The function may have other arguments, which
        will be passed down using `**kwargs`.
    task_name : str, optional
        Optional task name to use in :func:`asyncio.create_task`.
        By default, ``None``.
    **kwargs
        Keyword-argument pairs to pass to `validation_coroutine`. This
        should not include the document instance itself, which will be
        independently passed in as the first argument.

    Returns
    -------
    list of :class:`elm.web.document.BaseDocument`
        List of documents that passed the validation check, sorted by
        text length, with PDF documents taking the highest precedence.
    """
    searchers = [
        asyncio.create_task(
            validation_coroutine(doc, **kwargs), name=task_name
        )
        for doc in documents
    ]
    output = await asyncio.gather(*searchers)
    filtered_docs = [doc for doc, check in zip(documents, output) if check]
    return sorted(
        filtered_docs,
        key=lambda doc: (not isinstance(doc, PDFDocument), len(doc.text)),
    )


class PWKwargs:
    """Class to compile Playwright launch and context arguments"""

    _PE = PlaywrightEngine(stealth=True)
    SKIP_SCRIPS = []
    """List of scrapling stealth script names to skip"""

    USE_REALISTIC_VIEWPORTS = True
    """bool: Use realistic viewport sizes for the browser context"""

    _VIEWPORTS = {
        "desktop": [
            {"width": 1080, "height": 600, "device_scale_factor": 1},
            {"width": 1280, "height": 1024, "device_scale_factor": 1},
            {"width": 1366, "height": 768, "device_scale_factor": 1},
            {"width": 1440, "height": 900, "device_scale_factor": 1},
            {"width": 1536, "height": 864, "device_scale_factor": 1},
            {"width": 1600, "height": 900, "device_scale_factor": 1},
            {"width": 1680, "height": 1050, "device_scale_factor": 1},
            {"width": 1920, "height": 1080, "device_scale_factor": 1},
            {"width": 2560, "height": 1440, "device_scale_factor": 1.25},
            {"width": 3840, "height": 2160, "device_scale_factor": 2},
        ],
        "mobile": [
            # iPhone 6/7/8
            {"width": 375, "height": 667, "device_scale_factor": 2},
            # iPhone X/11 Pro
            {"width": 375, "height": 812, "device_scale_factor": 3},
            # iPhone 12/13/14
            {"width": 390, "height": 844, "device_scale_factor": 3},
            # iPhone XR/11
            {"width": 414, "height": 896, "device_scale_factor": 2},
            # iPhone 14 Pro Max
            {"width": 428, "height": 926, "device_scale_factor": 3},
            # Galaxy S9
            {"width": 360, "height": 800, "device_scale_factor": 2.5},
            # Pixel 7 Pro
            {"width": 412, "height": 915, "device_scale_factor": 2.625},
            # Pixel 6
            {"width": 393, "height": 873, "device_scale_factor": 2.75},
            # Pixel 5a
            {"width": 360, "height": 780, "device_scale_factor": 3},
        ],
        "tablet": [
            # iPad portrait
            {"width": 768, "height": 1024, "device_scale_factor": 2},
            # iPad landscape
            {"width": 1024, "height": 768, "device_scale_factor": 2},
            # Android tablet
            {"width": 800, "height": 1280, "device_scale_factor": 1.5},
            # Samsung Tab
            {"width": 962, "height": 601, "device_scale_factor": 1.5},
        ]
    }

    @classmethod
    def launch_kwargs(cls):
        """dict: kwargs to use for `playwright.chromium.launch()`"""
        return deepcopy(cls._PE._PlaywrightEngine__launch_kwargs())

    @classmethod
    def context_kwargs(cls, browser_type, ignore_https_errors=False):
        """dict: kwargs to use for `browser.new_context()`"""
        ck = deepcopy(cls._PE._PlaywrightEngine__context_kwargs())

        logger.trace("Loading browser context for browser type %r",
                     browser_type)

        platforms = (["desktop"]
                     if cls.USE_REALISTIC_VIEWPORTS
                     else ["desktop", "mobile", "tablet"])
        ua_info = (UserAgent(browsers=[browser_type], platforms=platforms)
                   .getBrowser("random"))
        logger.trace("User agent is:\n\t- %s", ua_info["useragent"])
        platform_type = ua_info["type"].casefold()

        if cls.USE_REALISTIC_VIEWPORTS:
            vp = deepcopy(choice(cls._VIEWPORTS[platform_type]))
            dsf = vp.pop("device_scale_factor")
        else:
            vp = {"width": randint(800, 1400), "height": randint(800, 1400)}
            dsf = choice([1, 1.25, 1.5, 2, 2.5, 2.625, 2.75, 3])
        logger.trace("Screen size is:\n\t- %r", vp)

        ck.update({"base_url": "http://127.0.0.1:443",
                   "device_scale_factor": dsf,
                   "extra_http_headers": DEFAULT_HEADERS,
                   "user_agent": ua_info["useragent"],
                   "viewport": vp,
                   "screen": vp,
                   "is_mobile": platform_type in {"mobile", "tablet"},
                   "has_touch": platform_type in {"mobile", "tablet"},
                   "ignore_https_errors": ignore_https_errors})
        return ck

    @classmethod
    def stealth_scripts(cls):
        """iterator: Iterator of scrapling scripts to use for stealth"""
        scripts = deepcopy(cls._PE._PlaywrightEngine__stealth_scripts())
        for script in scripts:
            if any(name in script for name in cls.SKIP_SCRIPS):
                continue
            yield script



"""
setup.py
"""
import os
from codecs import open
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, "README.rst"), encoding="utf-8") as f:
    readme = f.read()

with open(os.path.join(here, "elm", "version.py"), encoding="utf-8") as f:
    version = f.read()

version = version.split('=')[-1].strip().strip('"').strip("'")


with open("requirements.txt") as f:
    install_requires = f.readlines()


test_requires = ["pytest>=5.2", "pytest-mock", "pytest-asyncio", "pytest-cov",
                 "flaky>=3.8.1"]
description = "Energy Language Model"

setup(
    name="NREL-elm",
    version=version,
    description=description,
    long_description=readme,
    author="Grant Buster",
    author_email="grant.buster@nrel.gov",
    url="https://github.com/NREL/elm",
    packages=find_packages(),
    package_dir={"elm": "elm"},
    license="BSD 3-Clause",
    zip_safe=False,
    keywords="elm",
    python_requires='>=3.9',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    install_requires=install_requires,
    extras_require={
        "dev": install_requires + test_requires,
    },
    entry_points={"console_scripts": ["elm=elm.cli:main"]}
)



# -*- coding: utf-8 -*-
"""Custom Exceptions and Errors for ELM. """


class ELMError(Exception):
    """Generic ELM Error."""


class ELMKeyError(ELMError, KeyError):
    """ELM Key Error."""


class ELMInputError(ELMError, ValueError):
    """ELM Input (Value) Error."""


class ELMRuntimeError(ELMError, RuntimeError):
    """ELM RuntimeError."""
